import pathlib
import sys
import shutil
import os
import argparse
import json

from create_md import create_md

# {scheme_name}
# 	├── {amplicon_size}
# 	│   └── {version}
# 	│       ├── primer.bed
# 	│       ├── amplicon.bed
# 	│       ├── reference_msa.fasta
# 	│       ├── README.md
# 	│       └── work
# 	│	        ├── config.json
# 	│	        ├── scheme.png
# 	│	        ├── scheme.html
# 	│	        └── output.log
# 	├── LICENSE
# 	└── README.md


LICENSE_TXT = """------------------------------------------------------------------------

This work is licensed under a [Creative Commons Attribution-ShareAlike 4.0 International License](http://creativecommons.org/licenses/by-sa/4.0/) 

![](https://i.creativecommons.org/l/by-sa/4.0/88x31.png)"""


def cli():
    description = "Generates a github repo for a primerscheme"
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument(
        "--schemedir",
        help="The top level of the scheme",
        type=pathlib.Path,
        required=True,
        nargs="+",
    )
    parser.add_argument(
        "--schemeoutdir",
        help="Where to save and init the scheme repo",
        type=pathlib.Path,
        # required=True,
        default=pathlib.Path("/Users/kentcg/primerschemes"),
    )
    parser.add_argument(
        "--mdoutdir",
        help="The _pages dir of the github pages repo",
        type=pathlib.Path,
        default=pathlib.Path("/Users/kentcg/ChrisgKent.github.io/docs/_posts"),
    )
    parser.add_argument(
        "--dryrun",
        help="The _pages dir of the github pages repo",
        type=bool,
        default=True,
    )
    parser.add_argument(
        "--version",
        help="The Scheme version",
        type=str,
        default="v0.1.0",
    )
    parser.add_argument(
        "--size",
        help="The Scheme version",
        type=int,
        default=1000,
    )
    parser.add_argument(
        "--pclass",
        help="The Scheme version",
        choices=["primerschemes", "primerpanels"],
        required=True,
    )

    args = parser.parse_args()
    return args


def create(
    scheme_outdir: pathlib.Path,
    md_outdir,
    primal_nf_dir,
    size,
    version,
    scheme_name,
    pclass,
):
    # Check if already exists in outdir/index.json
    with open(scheme_outdir / "index.json", "r") as index:
        index_json = json.load(index)

        pclass_json = index_json[pclass]

        if scheme_name in pclass_json:
            if size in pclass_json[scheme_name]:
                if version in pclass_json[scheme_name][size]:
                    raise sys.exit(
                        f"{scheme_name} {size} {version} already exists in index.json"
                    )  # type: ignore

    primal_nf_scheme_dir = primal_nf_dir / primal_nf_dir.name
    common_name, taxid = primal_nf_dir.name.split("--")

    # Get paths
    primer_bed = primal_nf_scheme_dir / "output.primer.bed"
    conf = primal_nf_scheme_dir / "config.json"
    amplicon_bed = primal_nf_scheme_dir / "output.amplicon.bed"
    logfile = primal_nf_scheme_dir / "work/file.log"

    pngs: list[pathlib.Path] = [x for x in primal_nf_scheme_dir.glob("*.png")]
    htmls: list[pathlib.Path] = [x for x in primal_nf_scheme_dir.glob("*.html")]

    msa = primal_nf_dir / f"{taxid}_final.fasta"

    ## Create output dirs
    output_dir: pathlib.Path = scheme_outdir / f"{pclass}/{scheme_name}"
    size_dir: pathlib.Path = pathlib.Path(output_dir / size)
    version_dir: pathlib.Path = pathlib.Path(size_dir / version)

    # Check all the files exist
    if not primer_bed.exists():
        return None
    if not conf.exists():
        return None
    if not amplicon_bed.exists():
        return None
    if not logfile.exists():
        return None
    if not msa.exists():
        return None
    if len(pngs) == 0:
        return None
    if len(htmls) == 0:
        return None
    if not primal_nf_scheme_dir.exists():
        return None

    # If the output dir doesn't exist
    if not output_dir.exists():
        output_dir.mkdir()

        # Create the main readme
        with open(output_dir / "README.md", "w") as mainreadme:
            mainreadme.write(f"# PrimerSchemes for {common_name}\n\n")
            mainreadme.write(
                f"These are autogenerated schemes for TAXID: [{taxid}](https://www.ncbi.nlm.nih.gov/Taxonomy/Browser/wwwtax.cgi?mode=Info&id={taxid}&lvl=3&lin=f&keep=1&srchmode=1&unlock)\n\nGenomes are scraped from the EBI API, and then filtered, phylogenetically downsampled and aligned, to produce the input for a new PrimalScheme algorithm\n\nFor details please see preprint: TBD\n\n"
            )
            mainreadme.write(LICENSE_TXT)

        # Copy the licenseassets/LICENSE
        shutil.copyfile(pathlib.Path("assets/LICENSE"), output_dir / "LICENSE")

    # Create the size dir if doesn't exist
    if not size_dir.exists():
        size_dir.mkdir()

    # Create the version dir if doesn't exist
    if not version_dir.exists():
        version_dir.mkdir()
    else:
        return None

    ## Copy top level files
    # Copy the primer.bed file
    shutil.copyfile(primer_bed, version_dir / "primer.bed")
    # Copy the amplicon.bed file
    shutil.copyfile(amplicon_bed, version_dir / "amplicon.bed")
    # Copy the msa
    shutil.copyfile(msa, version_dir / msa.name)
    # Copy the licenseassets/LICENSE
    shutil.copyfile(pathlib.Path("assets/LICENSE"), version_dir / "LICENSE")

    ## Copy accessory filess
    # Create a work dir
    work_dir = version_dir / "work"
    work_dir.mkdir()

    # copy the config file
    shutil.copyfile(conf, work_dir / "config.json")
    # copy the logfile
    shutil.copyfile(logfile, work_dir / "output.log")

    # copy the pngs
    for png in pngs:
        shutil.copyfile(png, work_dir / png.name)

    # Create a scheme README.md
    with open(version_dir / "README.md", "w") as scheme_readme:
        scheme_readme.write(f"# {primal_nf_scheme_dir.name} {version}\n\n")

        ## Add the interactive links
        scheme_readme.write(
            f"### [Interactive page](https://chrisgkent.github.io/schemes/{primal_nf_dir.name}-{size}-{version})\n\n"
        )

        # Add a quick descriptions
        scheme_readme.write(f"### Description\n\n")
        scheme_readme.write(
            f"This is an autogenerated scheme for TAXID: [{taxid}](https://www.ncbi.nlm.nih.gov/Taxonomy/Browser/wwwtax.cgi?mode=Info&id={taxid}&lvl=3&lin=f&keep=1&srchmode=1&unlock)\n\nGenomes are scraped from the EBI API, and then filtered, phylogenetically downsampled and aligned, to produce the input for a new PrimalScheme algorithm\n\nFor details please see preprint: TBD\n\n"
        )

        scheme_readme.write(f"### Scheme Overview\n\n")
        # Add links to the pngs
        for png in pngs:
            scheme_readme.write(f"![Alt text](work/{png.name} '{png.name}')\n\n")

        # Add the footer
        scheme_readme.write(LICENSE_TXT)

    # Create the md page for jekyll
    md_path = create_md(
        html_figs=htmls,
        bed_file=primer_bed,
        config=conf,
        outdir=md_outdir,
        scheme_name=primal_nf_scheme_dir.name,
        url=f"{primal_nf_dir.name}-{size}-{version}",
    )


def main():
    args = cli()
    dryrun = args.dryrun
    # Get the outdirs
    md_outdir: pathlib.Path = args.mdoutdir
    scheme_outdir = args.schemeoutdir
    # get the input dir
    primal_nf_dirs = args.schemedir
    # get some params
    size = str(args.size)
    version = args.version

    for primal_nf_dir in primal_nf_dirs:
        scheme_name = primal_nf_dir.name
        create(
            scheme_outdir=scheme_outdir,
            md_outdir=md_outdir,
            primal_nf_dir=primal_nf_dir,
            size=size,
            version=version,
            scheme_name=scheme_name,
            pclass=args.pclass,
        )


if __name__ == "__main__":
    main()



import typer
import pathlib
from typing_extensions import Annotated
import shutil
import hashlib
import json
from typing import Optional
from enum import Enum
import requests

from primal_page.build_index import create_index
from primal_page.schemas import (
    PrimerClass,
    SchemeStatus,
    Info,
    determine_bedfile_version,
    BedfileVersion,
    validate_bedfile,
    BEDFILERESULT,
    Collection,
)


class FindResult(Enum):
    NOT_FOUND = 1
    FOUND = 2


# Create the typer app
app = typer.Typer()
modify_app = typer.Typer()
app.add_typer(
    modify_app,
    name="modify",
    help="Modify an existing scheme's metadata (info.json)",
)


LICENSE_TXT_CC_BY_SA_4_0 = """\n\n------------------------------------------------------------------------

This work is licensed under a [Creative Commons Attribution-ShareAlike 4.0 International License](http://creativecommons.org/licenses/by-sa/4.0/) 

![](https://i.creativecommons.org/l/by-sa/4.0/88x31.png)"""


def trim_file_whitespace(in_path: pathlib.Path, out_path: pathlib.Path):
    """
    Trim whitespace from the ends of a file.
        - Reads file into memory. Not suitable for large files
    """
    with open(in_path, "r") as infile:
        input_file = infile.read().strip()

    with open(out_path, "w") as outfile:
        outfile.write(input_file)


def regenerate_readme(path: pathlib.Path, info: Info, pngs: list[pathlib.Path]):
    """
    Regenerate the README.md file for a scheme

    :param path: The path to the scheme directory
    :type path: pathlib.Path
    :param info: The scheme information
    :type info: Info
    :param pngs: The list of PNG files
    :type pngs: list[pathlib.Path]
    """

    with open(path / "README.md", "w") as readme:
        readme.write(
            f"# {info.schemename} {info.ampliconsize}bp {info.schemeversion}\n\n"
        )

        if info.description != None:
            readme.write(f"## Description\n\n")
            readme.write(f"{info.description}\n\n")

        readme.write(f"## Overviews\n\n")
        for png in pngs:
            readme.write(f"![{png.name}](work/{png.name})\n\n")

        readme.write(f"## Details\n\n")

        # Write the detials into the readme
        readme.write(f"""```json\n{info.model_dump_json(indent=4)}\n```\n\n""")

        if info.license == "CC BY-SA 4.0":
            readme.write(LICENSE_TXT_CC_BY_SA_4_0)


def hashfile(fname: pathlib.Path) -> str:
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def find_ref(
    cli_reference: pathlib.Path | None,
    found_files: list[pathlib.Path],
    schemepath: pathlib.Path,
) -> pathlib.Path:
    """
    Find the reference.fasta file
    :param cli_reference: The reference.fasta file specified by the user. None if not specified
    :param found_files: A list of all files found in the scheme directory
    :param schemepath: The path to the scheme directory
    :return: The path to the reference.fasta file
    :raises FileNotFoundError: If the reference.fasta file cannot be found
    """
    # Search for reference.fasta
    if cli_reference is None:  # No reference specified
        # Search for a single *.fasta
        reference_list: list[pathlib.Path] = [
            path
            for path in found_files
            if path.name == ("reference.fasta") or path.name == ("referance.fasta")
        ]
        if len(reference_list) == 1:
            return reference_list[0]
        else:
            raise FileNotFoundError(
                f"Could not find a SINGLE reference.fasta file in {schemepath} or its subdirectories, found {len(reference_list)}. Please specify manually with --reference"
            )
    elif cli_reference.exists():
        # TODO validate the reference.fasta file
        return cli_reference
    else:
        raise FileNotFoundError(f"Could not find file at {cli_reference}")


def find_primerbed(
    cli_primerbed: pathlib.Path | None,
    found_files: list[pathlib.Path],
    schemepath: pathlib.Path,
) -> pathlib.Path:
    """
    Find the primer.bed file
    :param cli_primerbed: The primer.bed file specified by the user. None if not specified
    :param found_files: A list of all files found in the scheme directory
    :param schemepath: The path to the scheme directory
    :return: The path to the primer.bed file
    :raises FileNotFoundError: If the primer.bed file cannot be found
    """
    # Search for primer.bed
    if cli_primerbed is None:  # No primer.bed specified
        # Search for a single *.primer.bed
        primer_bed_list: list[pathlib.Path] = [
            path for path in found_files if path.name.endswith("primer.bed")
        ]
        if len(primer_bed_list) == 1:
            return primer_bed_list[0]
        else:
            raise FileNotFoundError(
                f"Could not find a SINGLE *.primer.bed file in {schemepath} or its subdirectories, found {len(primer_bed_list)}. Please specify manually with --primerbed"
            )
    elif cli_primerbed.exists():
        return cli_primerbed
    else:
        raise FileNotFoundError(f"Could not find file at {cli_primerbed}")


def find_config(
    cli_config: pathlib.Path | None,
    found_files: list[pathlib.Path],
    schemepath: pathlib.Path,
) -> tuple[FindResult, pathlib.Path | None]:
    """
    Find the config.json file
    :param cli_config: The config.json file specified by the user. None if not specified
    :param found_files: A list of all files found in the scheme directory
    :param schemepath: The path to the scheme directory
    :return: The path to the config.json file
    """
    # Search for config.json
    if cli_config is None:  # No config.json specified
        # Search for a single config.json
        config_list: list[pathlib.Path] = [
            path for path in found_files if path.name == ("config.json")
        ]
        match len(config_list):
            case 1:
                return (FindResult.FOUND, config_list[0])
            case _:
                return (FindResult.NOT_FOUND, None)

    elif cli_config.exists():
        # TODO validate the config.json file
        return (FindResult.FOUND, cli_config)
    else:
        return (FindResult.NOT_FOUND, None)


@app.command()
def create(
    schemepath: Annotated[
        pathlib.Path,
        typer.Argument(help="The path to the scheme directory", readable=True),
    ],
    schemename: Annotated[
        str,
        typer.Option(help="The name of the scheme"),
    ],
    ampliconsize: Annotated[
        int,
        typer.Option(help="Amplicon size", min=100),
    ],
    schemeversion: Annotated[
        str,
        typer.Option(
            help="Scheme version, default is parsed from config.json",
        ),
    ],
    species: Annotated[
        list[int],
        typer.Option(
            help="The species this scheme targets. Please use NCBI taxonomy ids"
        ),
    ],
    schemestatus: Annotated[
        SchemeStatus, typer.Option(help="Scheme status")
    ] = SchemeStatus.DRAFT.value,  # type: ignore
    citations: Annotated[
        list[str], typer.Option(help="Any associated citations. Please use DOI")
    ] = [],
    authors: Annotated[list[str], typer.Option(help="Any authors")] = [
        "quick lab",
        "artic network",
    ],
    primerbed: Annotated[
        Optional[pathlib.Path],
        typer.Option(
            help="Manually specify the primer bed file, default is *primer.bed",
            readable=True,
        ),
    ] = None,
    reference: Annotated[
        Optional[pathlib.Path],
        typer.Option(
            help="Manually specify the referance.fasta file, default is *.fasta",
            readable=True,
        ),
    ] = None,
    output: Annotated[
        pathlib.Path,
        typer.Option(help="Where to output the scheme", writable=True),
    ] = pathlib.Path("primerschemes"),
    configpath: Annotated[
        Optional[pathlib.Path],
        typer.Option(
            help="Where the config.json file is located", readable=True, exists=True
        ),
    ] = None,
    algorithmversion: Annotated[
        Optional[str], typer.Option(help="The version of primalscheme or other")
    ] = None,
    description: Annotated[
        Optional[str], typer.Option(help="A description of the scheme")
    ] = None,
    derivedfrom: Annotated[
        Optional[str],
        typer.Option(help="Which scheme has this scheme been derived from"),
    ] = None,
    primerclass: Annotated[
        PrimerClass, typer.Option(help="The primer class")
    ] = PrimerClass.PRIMERSCHEMES.value,  # type: ignore
    collection: Annotated[
        Optional[list[Collection]], typer.Option(help="The collection")
    ] = None,
):
    """Create a new scheme in the required format"""

    # Search for scheme repo for files
    found_files = [x for x in schemepath.rglob("*")]

    # Check for a single primer.bed file
    valid_primer_bed = find_primerbed(primerbed, found_files, schemepath)

    match validate_bedfile(valid_primer_bed):
        case BEDFILERESULT.VALID:
            pass
        case BEDFILERESULT.INVALID_VERSION:
            raise ValueError(
                f"Could not determine primerbed version for {valid_primer_bed}"
            )
        case BEDFILERESULT.INVALID_STRUCTURE:
            raise ValueError(
                f"Invalid primerbed structure for {valid_primer_bed}. Please ensure it is a valid 7 column bedfile"
            )
    # Get primerbed version
    primerbed_version: BedfileVersion = determine_bedfile_version(valid_primer_bed)

    # Find the reference.fasta file
    valid_ref = find_ref(reference, found_files, schemepath)

    # Search for config.json
    status, conf_path = find_config(configpath, found_files, schemepath)
    config_json: None | dict = None  # type: ignore
    if status == FindResult.FOUND and conf_path is not None:  # Second check is for mypy
        configpath = conf_path
        # Read in the config
        config_json: dict = json.load(configpath.open())
        # Remove some paths from the config
        config_json.pop("output_dir", None)
        # These hashes are no longer used, so remove to prevent confusion
        md5s_keys_to_remove = {k for k in config_json.keys() if k.endswith("md5")}
        for k in md5s_keys_to_remove:
            config_json.pop(k)

        if algorithmversion is None:
            if "algorithmversion" in config_json.keys():
                algorithmversion = str(config_json["algorithmversion"])
            else:
                raise ValueError(
                    f"algorithmversion not specified in {configpath}. Please specify manually with --algorithmversion"
                )

    elif status == FindResult.NOT_FOUND:
        if algorithmversion is None:
            raise FileNotFoundError(
                f"Could not find a config.json file in {schemepath}. Please specify manually with --configpath or specify algorithmversion with --algorithmversion"
            )

    # Multiple pngs/htmls/msas are allowed
    # The single check is mainly to prevent multiple schemes via providing the wrong directory
    # At this point we know we have a single scheme, due to having a single primer.bed, reference.fasta, and config.json

    # Search for pngs
    pngs = [path for path in found_files if path.name.endswith(".png")]
    # Search for html
    html = [path for path in found_files if path.name.endswith(".html")]
    # Search for msas
    msas = [
        path
        for path in found_files
        if path.name.endswith(".fasta") and path.name != "reference.fasta"
    ]

    # Copy all additional files to working directory
    # This is done to prevent preserve the original files
    misc_files_to_copy = [
        x
        for x in found_files
        if not x.name.endswith("primer.bed")
        and not x.name.endswith("config.json")
        and not x.name.endswith("info.json")
        and not x.name.endswith(".fasta")
        and not x.name.endswith(".png")
        and not x.name.endswith(".html")
        and not x.name.endswith(".db")  # Dont copy the mismatches db
        and x.name != ".DS_Store"  # Dont copy the macos file
        and x.is_file()
    ]

    # Create the collections set
    collections = {x for x in collection} if collection is not None else set()

    # Create the info.json
    # Generate the md5s
    info = Info(
        ampliconsize=ampliconsize,
        schemeversion=schemeversion,
        schemename=schemename,
        primer_bed_md5="NONE",  # Will be updated later
        reference_fasta_md5="NONE",  # Will be updated later
        status=schemestatus,
        citations=set(citations),
        authors=set(authors),
        algorithmversion=algorithmversion,  # type: ignore
        species=set(species),
        description=description,
        derivedfrom=derivedfrom,
        primerclass=primerclass,
        articbedversion=primerbed_version,
        collections=collections,
    )

    #####################################
    # Final validation and create files #
    #####################################
    # Everything needs to be validated before creating files

    # Check if the repo already exists
    repo_dir = output / schemename / str(ampliconsize) / schemeversion
    if repo_dir.exists():
        raise FileExistsError(f"{repo_dir} already exists")
    repo_dir.mkdir(parents=True)

    # If this fails it will deleted the half completed scheme
    # Need to check the repo doesnt already exist
    try:
        # Copy files and trim whitespace
        trim_file_whitespace(valid_primer_bed, repo_dir / "primer.bed")
        trim_file_whitespace(valid_ref, repo_dir / "reference.fasta")

        # Update the hashes in the info.json
        info.primer_bed_md5 = hashfile(repo_dir / "primer.bed")
        info.reference_fasta_md5 = hashfile(repo_dir / "reference.fasta")

        working_dir = repo_dir / "work"
        working_dir.mkdir()

        if config_json is not None:
            # Write out the config.json
            with open(working_dir / "config.json", "w") as configfile:
                json.dump(config_json, configfile, indent=4, sort_keys=True)

        # Copy over misc files
        for misc_file in misc_files_to_copy:
            shutil.copy(misc_file, working_dir / misc_file.name)

        # Copy the pngs
        for png in pngs:
            shutil.copy(png, working_dir / png.name)
        # Copy the htmls
        for html in html:
            shutil.copy(html, working_dir / html.name)
        # Copy the msas
        for msa in msas:
            shutil.copy(msa, working_dir / msa.name)

        # Write info.json
        with open(repo_dir / "info.json", "w") as infofile:
            infofile.write(info.model_dump_json(indent=4))

        # Create a README.md with link to all pngs
        regenerate_readme(repo_dir, info, pngs)
    except Exception as e:
        # Cleanup
        shutil.rmtree(repo_dir)
        raise Exception(f"{e}\nCleaning up {repo_dir}")


@modify_app.command()
def status(
    schemeinfo: Annotated[
        pathlib.Path,
        typer.Argument(help="The path to info.json", readable=True, exists=True),
    ],
    schemestatus: Annotated[
        SchemeStatus,
        typer.Option(
            help="The scheme class",
        ),
    ] = SchemeStatus.DRAFT,
):
    """Change the status field in the info.json"""

    info = json.load(schemeinfo.open())
    info = Info(**info)

    if info.status == schemestatus.value:
        raise ValueError(f"{schemeinfo} status is already {schemestatus}")
    else:
        info.status = schemestatus

    # Write the validated info.json
    with open(schemeinfo, "w") as infofile:
        infofile.write(info.model_dump_json(indent=4))

    # Update the README
    scheme_path = schemeinfo.parent
    pngs = [path for path in scheme_path.rglob("*.png")]
    regenerate_readme(scheme_path, info, pngs)


@modify_app.command()
def primerclass(
    schemeinfo: Annotated[
        pathlib.Path,
        typer.Argument(help="The path to info.json", readable=True, exists=True),
    ],
    primerclass: Annotated[
        PrimerClass, typer.Argument(help="The primerclass to change to")
    ],
):
    """Append an author to the authors list in the info.json file"""

    info = json.load(schemeinfo.open())
    info = Info(**info)

    # Check if author is already in the list
    info.primerclass = primerclass

    # Write the validated info.json
    with open(schemeinfo, "w") as infofile:
        infofile.write(info.model_dump_json(indent=4))

    # Update the README
    scheme_path = schemeinfo.parent
    pngs = [path for path in scheme_path.rglob("*.png")]
    regenerate_readme(scheme_path, info, pngs)


@modify_app.command()
def add_author(
    schemeinfo: Annotated[
        pathlib.Path,
        typer.Argument(help="The path to info.json", readable=True, exists=True),
    ],
    author: Annotated[str, typer.Argument(help="The author to add")],
):
    """Append an author to the authors list in the info.json file"""

    info = json.load(schemeinfo.open())
    info = Info(**info)

    # Check if author is already in the list
    if author in info.authors:
        raise ValueError(f"{author} is already in the authors list")
    info.authors.add(author)

    # Write the validated info.json
    with open(schemeinfo, "w") as infofile:
        infofile.write(info.model_dump_json(indent=4))

    # Update the README
    scheme_path = schemeinfo.parent
    pngs = [path for path in scheme_path.rglob("*.png")]
    regenerate_readme(scheme_path, info, pngs)


@modify_app.command()
def remove_author(
    schemeinfo: Annotated[
        pathlib.Path,
        typer.Argument(help="The path to info.json", readable=True, exists=True),
    ],
    author: Annotated[str, typer.Argument(help="The author to remove")],
):
    """Remove an author from the authors list in the info.json file"""
    info = json.load(schemeinfo.open())
    info = Info(**info)

    # Check if author is already not in the list
    if author not in info.authors:
        raise ValueError(f"{author} is already not in the authors list")
    info.authors.remove(author)

    # Write the validated info.json
    with open(schemeinfo, "w") as infofile:
        infofile.write(info.model_dump_json(indent=4))

    # Update the README
    scheme_path = schemeinfo.parent
    pngs = [path for path in scheme_path.rglob("*.png")]
    regenerate_readme(scheme_path, info, pngs)


@modify_app.command()
def add_citation(
    schemeinfo: Annotated[
        pathlib.Path,
        typer.Argument(help="The path to info.json", readable=True, exists=True),
    ],
    citation: Annotated[str, typer.Argument(help="The citation to add")],
):
    """Append an citation to the authors list in the info.json file"""
    info = json.load(schemeinfo.open())
    info = Info(**info)

    # Check if citation is already in the list
    if citation in info.citations:
        raise ValueError(f"{citation} is areadly in the citation list")
    info.citations.add(citation)

    # Write the validated info.json
    with open(schemeinfo, "w") as infofile:
        infofile.write(info.model_dump_json(indent=4))

    # Update the README
    scheme_path = schemeinfo.parent
    pngs = [path for path in scheme_path.rglob("*.png")]
    regenerate_readme(scheme_path, info, pngs)


@modify_app.command()
def remove_citation(
    schemeinfo: Annotated[
        pathlib.Path,
        typer.Argument(help="The path to info.json", readable=True, exists=True),
    ],
    citation: Annotated[str, typer.Argument(help="The citation to remove")],
):
    """Remove an citation form the authors list in the info.json file"""
    info = json.load(schemeinfo.open())
    info = Info(**info)

    if citation not in info.citations:
        raise ValueError(f"{citation} is not in the citation list")
    info.citations.remove(citation)

    # Write the validated info.json
    with open(schemeinfo, "w") as infofile:
        infofile.write(info.model_dump_json(indent=4))

    # Update the README
    scheme_path = schemeinfo.parent
    pngs = [path for path in scheme_path.rglob("*.png")]
    regenerate_readme(scheme_path, info, pngs)


@modify_app.command()
def remove_collection(
    schemeinfo: Annotated[
        pathlib.Path,
        typer.Argument(help="The path to info.json", readable=True, exists=True),
    ],
    collection: Annotated[Collection, typer.Argument(help="The Collection to remove")],
):
    """Remove an Collection from the Collection list in the info.json file"""
    info = json.load(schemeinfo.open())
    info = Info(**info)

    # Check if collection is already not in the list
    if collection not in info.collections:
        raise ValueError(f"{collection} is already not in the collection list")
    info.collections.remove(collection)

    # Write the validated info.json
    with open(schemeinfo, "w") as infofile:
        infofile.write(info.model_dump_json(indent=4))

    # Update the README
    scheme_path = schemeinfo.parent
    pngs = [path for path in scheme_path.rglob("*.png")]
    regenerate_readme(scheme_path, info, pngs)


@modify_app.command()
def add_collection(
    schemeinfo: Annotated[
        pathlib.Path,
        typer.Argument(help="The path to info.json", readable=True, exists=True),
    ],
    collection: Annotated[Collection, typer.Argument(help="The Collection to add")],
):
    """Add a Collection to the Collection list in the info.json file"""
    info = json.load(schemeinfo.open())
    info = Info(**info)

    # Check if author is already not in the list
    if collection in info.collections:
        raise ValueError(f"{collection} is already in the collection list")
    info.collections.add(collection)

    # Write the validated info.json
    with open(schemeinfo, "w") as infofile:
        infofile.write(info.model_dump_json(indent=4))

    # Update the README
    scheme_path = schemeinfo.parent
    pngs = [path for path in scheme_path.rglob("*.png")]
    regenerate_readme(scheme_path, info, pngs)


@modify_app.command()
def description(
    schemeinfo: Annotated[
        pathlib.Path,
        typer.Argument(help="The path to info.json", readable=True, exists=True),
    ],
    description: Annotated[
        str,
        typer.Argument(
            help="The new description. Use 'None' to remove the description"
        ),
    ],
):
    """Replaces the description in the info.json file"""
    info = json.load(schemeinfo.open())
    info = Info(**info)

    # Add the description
    if description == "None":
        info.description = None
    else:
        info.description = description.strip()

    # Write the validated info.json
    with open(schemeinfo, "w") as infofile:
        infofile.write(info.model_dump_json(indent=4))

    # Update the README
    scheme_path = schemeinfo.parent
    pngs = [path for path in scheme_path.rglob("*.png")]
    regenerate_readme(scheme_path, info, pngs)


@modify_app.command()
def derivedfrom(
    schemeinfo: Annotated[
        pathlib.Path,
        typer.Argument(help="The path to info.json", readable=True, exists=True),
    ],
    derivedfrom: Annotated[
        str,
        typer.Argument(
            help="The new derivedfrom. Use 'None' to remove the derivedfrom"
        ),
    ],
):
    """Replaces the derivedfrom in the info.json file"""
    info = json.load(schemeinfo.open())
    info = Info(**info)

    # Add the derivedfrom
    if derivedfrom == "None":
        info.derivedfrom = None
    else:
        info.derivedfrom = derivedfrom.strip()

    # Write the validated info.json
    with open(schemeinfo, "w") as infofile:
        infofile.write(info.model_dump_json(indent=4))

    # Update the README
    scheme_path = schemeinfo.parent
    pngs = [path for path in scheme_path.rglob("*.png")]
    regenerate_readme(scheme_path, info, pngs)


@modify_app.command()
def license(
    schemeinfo: Annotated[
        pathlib.Path,
        typer.Argument(help="The path to info.json", readable=True, exists=True),
    ],
    license: Annotated[
        str,
        typer.Argument(
            help="The new license. Use 'None' show the work is not licensed (Not recommended)"
        ),
    ],
):
    """Replaces the license in the info.json file"""
    info = json.load(schemeinfo.open())
    info = Info(**info)

    info.license = license.strip()

    # Write the validated info.json
    with open(schemeinfo, "w") as infofile:
        infofile.write(info.model_dump_json(indent=4))

    # Update the README
    scheme_path = schemeinfo.parent
    pngs = [path for path in scheme_path.rglob("*.png")]
    regenerate_readme(scheme_path, info, pngs)


@app.command()
def build_index(
    gitaccount: Annotated[
        str,
        typer.Option(help="The name of the github account"),
    ] = "quick-lab",
    gitserver: Annotated[
        str,
        typer.Option(help="The name of the github server"),
    ] = "https://github.com/",
    parentdir: Annotated[
        pathlib.Path, typer.Option(help="The parent directory")
    ] = pathlib.Path("."),
    git_commit_sha: Annotated[
        Optional[str], typer.Option(help="The git commit")
    ] = None,
    force: Annotated[
        bool, typer.Option(help="Force the creation of the index.json")
    ] = False,
):
    """Build an index.json file from all schemes in the directory"""

    create_index(gitserver, gitaccount, parentdir, git_commit_sha, force)


@app.command()
def remove(
    schemeinfo: Annotated[
        pathlib.Path,
        typer.Argument(help="The path to info.json", readable=True, exists=True),
    ],
):
    """Remove a scheme's version from the repo, will also remove size and schemename directories if empty"""
    # Check that this is an info.json file (for safety)
    if schemeinfo.name != "info.json":
        raise ValueError(f"{schemeinfo} is not an info.json file")

    # Remove the schemeversion directory
    shutil.rmtree(schemeinfo.parent)

    # Move up the path and remove the size directory if empty
    size_dir = schemeinfo.parent.parent
    if len(list(size_dir.iterdir())) == 0:
        size_dir.rmdir()

    # Move up the path and remove the schemename directory if empty
    scheme_dir = size_dir.parent
    if len(list(scheme_dir.iterdir())) == 0:
        scheme_dir.rmdir()


@app.command()
def regenerate(
    schemeinfo: Annotated[
        pathlib.Path,
        typer.Argument(help="The path to info.json", readable=True, exists=True),
    ],
):
    """
    Regenerate the info.json and README.md file for a scheme
        - Rehashes info.json's primer_bed_md5 and reference_fasta_md5
        - Regenerates the README.md file
        - Recalculate the artic-primerbed version

    Ensures work/config.json has no absolute paths
        - Ensures hashes in config.json are removed
    """
    # Check that this is an info.json file (for safety)
    if schemeinfo.name != "info.json":
        raise ValueError(f"{schemeinfo} is not an info.json file")

    # Get the scheme path
    scheme_path = schemeinfo.parent

    # Get the info
    info_json = json.load(schemeinfo.open())

    # Trim whitespace from primer.bed and reference.fasta
    trim_file_whitespace(scheme_path / "primer.bed", scheme_path / "primer.bed")
    trim_file_whitespace(
        scheme_path / "reference.fasta", scheme_path / "reference.fasta"
    )

    # if articbedversion not set then set it
    articbedversion = determine_bedfile_version(scheme_path / "primer.bed")
    if articbedversion == BedfileVersion.INVALID:
        raise ValueError(
            f"Could not determine artic-primerbed version for {scheme_path / 'primer.bed'}"
        )
    info_json["articbedversion"] = articbedversion.value

    # Regenerate the files hashes
    info_json["primer_bed_md5"] = hashfile(scheme_path / "primer.bed")
    info_json["reference_fasta_md5"] = hashfile(scheme_path / "reference.fasta")

    info = Info(**info_json)

    # Get the pngs
    pngs = [path for path in scheme_path.rglob("*.png")]

    #####################################
    # Final validation and create files #
    #####################################

    # Write the validated info.json
    with open(schemeinfo, "w") as infofile:
        infofile.write(info.model_dump_json(indent=4))

    # Regenerate the readme
    regenerate_readme(scheme_path, info, pngs)


@app.command()
def download(
    output: Annotated[pathlib.Path, typer.Option(help="Where to output the schemes")],
    index_url: Annotated[
        str, typer.Option(help="The URL to the index.json")
    ] = "https://raw.githubusercontent.com/quick-lab/primerschemes/main/index.json",
):
    """Download all schemes from the index.json"""
    # Download the index.json
    index = json.loads(requests.get(index_url).text)

    # Create the output directory
    output_primerschemes = output / "primerschemes"
    output_primerschemes.mkdir(exist_ok=False)

    # Grab the primerschemes
    primerschemes = index.get("primerschemes", {})

    # Download all the schemes
    for schemename in primerschemes:
        for ampliconsize in primerschemes[schemename]:
            for schemeversion in primerschemes[schemename][ampliconsize]:
                scheme = primerschemes[schemename][ampliconsize][schemeversion]
                scheme_dir = (
                    output_primerschemes
                    / schemename
                    / str(ampliconsize)
                    / schemeversion
                )
                scheme_dir.mkdir(parents=True, exist_ok=True)

                # Download the bedfile
                bedfile_url = scheme["primer_bed_url"]
                bedfile_text = requests.get(bedfile_url).text
                bedfile_hash = hashlib.md5(bedfile_text.encode()).hexdigest()
                # Check hashes before writing
                if bedfile_hash != scheme["primer_bed_md5"]:
                    raise ValueError(
                        f"Hash mismatch for {scheme['primer_bed_md5']}. Expected {scheme['primer_bed_md5']} but got {bedfile_hash}"
                    )
                # Write the file
                with open(scheme_dir / "primer.bed", "w") as f:
                    f.write(bedfile_text)

                # Download the reference
                reference_url = scheme["reference_fasta_url"]
                reference_text = requests.get(reference_url).text
                reference_hash = hashlib.md5(reference_text.encode()).hexdigest()
                # Check hashes before writing
                if reference_hash != scheme["reference_fasta_md5"]:
                    raise ValueError(
                        f"Hash mismatch for {scheme['reference_fasta_md5']}. Expected {scheme['reference_fasta_md5']} but got {reference_hash}"
                    )
                # Write the file
                with open(scheme_dir / "reference.fasta", "w") as f:
                    f.write(reference_text)

                # Download the info.json
                info_url = scheme["info_json_url"]
                info_text = requests.get(info_url).text
                # Write the file
                with open(scheme_dir / "info.json", "w") as f:
                    f.write(info_text)

                print(f"Downloaded:\t{schemename}/{ampliconsize}/{schemeversion}")


if __name__ == "__main__":
    app()



import argparse
import pathlib
import json
import sys
import datetime


def cli():
    description = "Generates a primerscheme from an MSA"
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument(
        "--schemedir",
        help="The top level of the scheme",
        type=pathlib.Path,
        required=True,
    )
    parser.add_argument(
        "--outdir",
        help="The top level of the scheme",
        type=pathlib.Path,
        required=True,
    )

    args = parser.parse_args()

    if not args.outdir.is_dir():
        raise sys.exit(f"{args.output} not a directory")  # type: ignore

    return args


def create_md(
    html_figs: list[pathlib.Path],
    bed_file: pathlib.Path,
    config: pathlib.Path,
    outdir: pathlib.Path,
    scheme_name: str,
    url: str,
):
    """
    Creates a markdown page for a primerscheme, for use with Jekyll.

    :param html_figs: A list of paths to the html figures.
    :param bed_file: A path to the bed file.
    :param config: A path to the config file.
    :param outdir: A path to the output directory.
    :param scheme_name: The name of the scheme.
    :param url: The full name of the scheme {scheme_name}-{len}-{version}.
    """

    # Create some data
    scheme_date = datetime.datetime.now().strftime("%Y-%m-%d")

    page_path = outdir / str(scheme_date + "-" + url + ".md")
    if page_path.exists():
        sys.exit(f"{page_path} already exists")

    version = "v0.0.1"

    ## Create the header string
    header = f"""---\nlayout: post\ntitle: "{scheme_name}"\ncategories: CATEGORY-1 CATEGORY-2\ndescription: PrimerScheme for {scheme_name} {version}\npermalink: /schemes/{url}/\n---\n\n"""

    # Write the file
    with open(page_path, "w") as page:
        # Write the header to the file
        page.write(header + "\n")

        # Write the Scheme figure
        for fig in html_figs:
            with open(fig, "r") as html_fig:
                page.write(html_fig.read() + "\n")

        page.write("## Bedfile\n")

        # Write the bed file
        page.write("```" + "\n")
        # Write the bed file
        with open(bed_file, "r") as bed:
            page.write(bed.read() + "\n")
        page.write("```" + "\n")

        # Write the config file
        config_json = json.load(open(config, "r"))

        page.write("## Config\n\n")
        page.write("```json" + "\n")
        page.write(json.dumps(config_json) + "\n")
        page.write("```" + "\n")

    return page_path



from pydantic import BaseModel, PositiveInt
from pydantic.functional_validators import AfterValidator
from typing import Annotated
import re
from enum import Enum
import pathlib
from primalhelper.bedfile import read_bed_file

SCHEMENAME_PATTERN = r"^[a-z0-9][a-z0-9-]*[a-z0-9]$"
VERSION_PATTERN = r"^v\d+\.\d+\.\d+$"

# Primername versions
V2_PRIMERNAME = r"^[a-zA-Z0-9\-]+_[0-9]+_(LEFT|RIGHT)_[0-9]+$"
V1_PRIMERNAME = r"^[a-zA-Z0-9\-]+_[0-9]+_(LEFT|RIGHT)(_ALT[0-9]*|_alt[0-9]*)*$"

# Bedfile versions
## This doesn't parse the contents just the structure
BEDFILE_LINE = r"^\S+\t\d+\t\d+\t\S+\t\d+\t(\+|\-)\t[a-zA-Z]+$"


class BEDFILERESULT(Enum):
    VALID = 0
    INVALID_VERSION = 1
    INVALID_STRUCTURE = 2


def validate_bedfile_line_structure(line: str) -> bool:
    line = line.strip()
    if line.startswith("#"):
        return True
    return re.search(BEDFILE_LINE, line) is not None


def validate_bedfile(bedfile: pathlib.Path) -> BEDFILERESULT:
    # Read in the bedfile string
    bedfile_str = bedfile.read_text()

    # Split the bedfile into lines
    bedlines = bedfile_str.split("\n")

    # Check each line
    for line in bedlines:
        if not validate_bedfile_line_structure(line):
            return BEDFILERESULT.INVALID_STRUCTURE

    # Check the bedfile names.
    match determine_bedfile_version(bedfile):
        case BedfileVersion.INVALID:
            return BEDFILERESULT.INVALID_VERSION
        case _:
            return BEDFILERESULT.VALID


# bedfile versions
class BedfileVersion(Enum):
    """
    V1 bedfiles use a 6 col system
    V2 bedfiles use a 7 col system and V1 primernames
    V3 bedfiles use a 7 col system and V2 primernames
    """

    V1 = "v1.0"
    V2 = "v2.0"
    V3 = "v3.0"
    INVALID = "invalid"  # Not applicable


def determine_bedfile_version(bedfile: pathlib.Path) -> BedfileVersion:
    """
    Determine the bedfile version
    :param bedfile: The bedfile to check
    :return: The bedfile version
    """
    bedlines, _headers = read_bed_file(bedfile)
    # If 6 cols then v1
    if len(bedlines[0]) == 6:
        return BedfileVersion.V1

    # If 7 cols then v2 or v3
    # Check from primername
    primernames = [x[3] for x in bedlines]
    primer_name_versions = {determine_primername_version(x) for x in primernames}
    if primer_name_versions == {PrimerNameVersion.V1}:
        return BedfileVersion.V2
    elif primer_name_versions == {PrimerNameVersion.V2}:
        return BedfileVersion.V3
    # Invalid if we get here
    # Mix of v1, v2 or invalid
    return BedfileVersion.INVALID


class PrimerNameVersion(Enum):
    V1 = "v1"
    V2 = "v2"
    INVALID = "invalid"  # Not applicable


def determine_primername_version(primername: str) -> PrimerNameVersion:
    """
    Determine the primername version
    :param primername: The primername to check
    :return: The primername version
    """
    if re.search(V2_PRIMERNAME, primername):
        return PrimerNameVersion.V2
    elif re.search(V1_PRIMERNAME, primername):
        return PrimerNameVersion.V1
    else:
        return PrimerNameVersion.INVALID


def convert_v1_primernames_to_v2(primername: str) -> str:
    """
    Convert a v1 primername to a v2 primername. Cannot handle alt primers
    :param primername: The v1 primername
    :return: The v2 primername
    """
    # Check if this is a v1 primername
    if determine_primername_version(primername) != PrimerNameVersion.V1:
        raise ValueError(f"{primername} is not a valid v1 primername")

    # Split the primername
    data = primername.split("_")
    # Remove the alt
    if data[-1] == "alt" or data[-1] == "ALT":
        raise ValueError(f"{primername} is a v1 alt primername, cannot convert")

    data.append("0")
    # Join back together
    return "_".join(data)


class PrimerClass(Enum):
    PRIMERSCHEMES = "primerschemes"


class SchemeStatus(Enum):
    WITHDRAWN = "withdrawn"
    DEPRECATED = "deprecated"
    AUTOGENERATED = "autogenerated"
    DRAFT = "draft"
    TESTING = "tested"
    VALIDATED = "validated"


def validate_schemeversion(version: str) -> str:
    if not re.match(VERSION_PATTERN, version):
        raise ValueError(
            f"Invalid version: {version}. Must match be in form of v(int).(int).(int)"
        )
    return version


def validate_schemename(schemename: str) -> str:
    if not re.match(SCHEMENAME_PATTERN, schemename):
        raise ValueError(
            f"Invalid schemename: {schemename}. Must only contain a-z, 0-9, and -. Cannot start or end with -"
        )
    return schemename


def not_empty(x: list | set) -> list | set:
    if len(x) == 0:
        raise ValueError("Cannot be empty")
    return x


class Collection(Enum):
    # Authors
    ARTIC = "ARTIC"
    MODJADJI = "MODJADJI"
    QUICK_LAB = "QUICK-LAB"
    COMMUNITY = "COMMUNITY"
    # Uses
    WASTE_WATER = "WASTE-WATER"
    CLINAL_ISOLATES = "CLINAL-ISOLATES"
    # Coverage types
    WHOLE_GENOME = "WHOLE-GENOME"
    PANEL = "PANEL"
    MULTI_TARGET = "MULTI-TARGET"


class Info(BaseModel):
    ampliconsize: PositiveInt
    schemeversion: Annotated[str, AfterValidator(validate_schemeversion)]
    schemename: Annotated[str, AfterValidator(validate_schemename)]
    primer_bed_md5: str
    reference_fasta_md5: str
    status: SchemeStatus
    citations: set[str]
    authors: Annotated[set[str], AfterValidator(not_empty)]
    algorithmversion: str
    species: Annotated[set[int | str], AfterValidator(not_empty)]
    license: str = "CC BY-SA 4.0"
    primerclass: PrimerClass = PrimerClass.PRIMERSCHEMES
    infoschema: str = "v1.3.0"
    articbedversion: BedfileVersion
    # Add the optional fields
    description: str | None = None
    derivedfrom: str | None = None
    collections: set[Collection] = set()


if __name__ == "__main__":
    info = Info(
        ampliconsize=400,
        schemeversion="v0.0.0",
        schemename="test",
        primer_bed_md5="hello",
        reference_fasta_md5="world",
        status=SchemeStatus.DRAFT,
        citations=set(),
        authors=set("artic"),
        algorithmversion="test",
        species=set("sars-cov-2"),
        articbedversion=BedfileVersion.V3,
        collections=set(),
    )

    # indexv = IndexVersion(
    #    ampliconsize=400,
    #    schemeversion="v0.0.0",
    #    schemename="test",
    #    primer_bed_md5="hello",
    #    reference_fasta_md5="world",
    #    status=SchemeStatus.DRAFT,
    #    citations=[],
    #    authors=[],
    #    algorithmversion="test",
    #    species=[1, "hello", 1],
    #    IndexVersion="test",
    # )

    print(info.model_dump_json(indent=4))

    info.authors.add("hello")
    print(info.model_dump_json(indent=4))

    # print(indexv.model_dump_json(indent=4))
